import os
from requests_cache import CachedSession

dirname = os.path.dirname(__file__)
CACHE_FILE_PATH = os.path.join(dirname, "requests_cache.sqlite")

requests_cache = CachedSession(allowable_methods=("GET", "POST"), cache_name=CACHE_FILE_PATH)



import json
from requests.utils import dict_from_cookiejar
from base64 import urlsafe_b64encode

from .requests_handler import requests_cache
from .helpers.utils import safely_get_value_from_key, get_auth_header, CLIENT_VERSION
from .reply import Reply


class Comment(object):
    FORMAT_URLS = {
        "POST": "https://www.youtube.com/post/{}",
        # HARD_CODED: This key seems to be constant to everyone, IDK
        "BROWSE_ENDPOINT": "https://www.youtube.com/youtubei/v1/browse?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
        "UPDATE_COMMENT_ENDPOINT": "https://www.youtube.com/youtubei/v1/comment/update_comment?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8&prettyPrint=false",
        "PERFORM_COMMENT_ACTION_ENDPOINT": "https://www.youtube.com/youtubei/v1/comment/perform_comment_action?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8&prettyPrint=false",
        "FIXED_COMMENT": "https://www.youtube.com/channel/{}/community?lc={}&lb={}",
    }

    def __init__(
        self,
        post_id,
        comment_id,
        channel_id=None,
        author=None,
        content_text=None,
        vote_count=None,
        replies_continuation_token=None,
        click_tracking_params=None,
        visitor_data=None,
        session_index="0",
    ):
        self.post_id = post_id
        self.comment_id = comment_id
        self.channel_id = channel_id
        self.author = author
        self.content_text = content_text
        self.vote_count = vote_count
        self.replies_continuation_token = replies_continuation_token
        self.click_tracking_params = click_tracking_params
        self.visitor_data = visitor_data
        self.session_index = session_index
        self.replies = []

    def as_json(self):
        return {
            "comment_id": self.comment_id,
            "post_id": self.post_id,
            "channel_id": self.channel_id,
            "author": self.author,
            "content_text": self.content_text,
            "vote_count": self.vote_count,
        }

    def __str__(self):
        return json.dumps(self.as_json(), indent=4)

    def __repr__(self):
        return self.__str__()

    def get_text(self):
        if self.content_text is not None:
            return "".join([run["text"] for run in self.content_text["runs"]])
        return None

    def load_replies(self, expire_after=0):
        headers = {
            "x-origin": "https://www.youtube.com",
            "Referer": Comment.FORMAT_URLS["POST"].format(self.post_id),
        }

        # Add authorization header
        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        if self.replies_continuation_token:
            headers.update(
                {
                    "X-Goog-AuthUser": self.session_index,
                    "X-Origin": "https://www.youtube.com",
                    "X-Youtube-Client-Name": "1",
                    "X-Youtube-Client-Version": CLIENT_VERSION,
                }
            )

            json_body = {
                "context": {
                    "client": {
                        "clientName": "WEB",
                        "clientVersion": CLIENT_VERSION,
                        "originalUrl": Comment.FORMAT_URLS["POST"].format(self.post_id),
                        "visitorData": self.visitor_data,
                    }
                },
                "continuation": self.replies_continuation_token,
                "clickTracking": {"clickTrackingParams": self.click_tracking_params},
            }

            r = requests_cache.post(Comment.FORMAT_URLS["BROWSE_ENDPOINT"], json=json_body, expire_after=expire_after, headers=headers)

            data = r.json()
            append = safely_get_value_from_key(data, "onResponseReceivedEndpoints", 0, "appendContinuationItemsAction", default=[])
            self.click_tracking_params = data["trackingParams"]
            continuation_items = safely_get_value_from_key(append, "continuationItems", default=[])

            self.append_replies_from_items(continuation_items)

    def append_replies_from_items(self, items):
        there_is_no_continuation_token = True
        for item in items:
            kind = list(item.keys())[0]

            if kind == "commentRenderer":
                self.replies.append(Reply.from_data(item[kind]))
            elif kind == "continuationItemRenderer":
                if "continuationEndpoint" in item[kind]:
                    self.replies_continuation_token = item[kind]["continuationEndpoint"]["continuationCommand"]["token"]
                    there_is_no_continuation_token = False
                elif "button" in item[kind]:
                    self.replies_continuation_token = item[kind]["button"]["buttonRenderer"]["command"]["continuationCommand"]["token"]
                    there_is_no_continuation_token = False

        if there_is_no_continuation_token:
            self.replies_continuation_token = False

    @staticmethod
    def from_data(data, post_id, channel_id, replies_continuation_token, click_tracking_params, visitor_data, session_index):
        comment = Comment(
            post_id,
            data["commentId"],
            channel_id=channel_id,
            content_text=safely_get_value_from_key(data, "contentText"),
            author={
                "authorText": safely_get_value_from_key(data, "authorText"),
                "authorThumbnail": safely_get_value_from_key(data, "authorThumbnail"),
                "authorEndpoint": safely_get_value_from_key(data, "authorEndpoint", "browseEndpoint"),
                "authorIsChannelOwner": safely_get_value_from_key(data, "authorIsChannelOwner"),
                "sponsorCommentBadge": safely_get_value_from_key(data, "sponsorCommentBadge"),
            },
            vote_count=safely_get_value_from_key(data, "voteCount"),
            replies_continuation_token=replies_continuation_token,
            click_tracking_params=click_tracking_params,
            visitor_data=visitor_data,
            session_index=session_index,
        )

        comment.raw_data = data

        return comment

    @staticmethod
    def get_fixed_comment_params(comment_id, post_id, channel_id):
        part1 = [
            b"\x12\tcommunity\xB8\x01\x00\xCA\x01",
            (32 + len(post_id)).to_bytes(1, "big"),
            b"\x82\x01",
            len(comment_id).to_bytes(1, "big"),
            comment_id.encode(),
            b"\xB2\x01",
            len(post_id).to_bytes(1, "big"),
            post_id.encode(),
            b"\xEA\x02\x04\x10\x01\x18\x01\xAA\x03",
            (84 + len(post_id)).to_bytes(1, "big"),
            b"\x22",
            (64 + len(post_id)).to_bytes(1, "big"),
            b"0\x00\x82\x01",
            len(comment_id).to_bytes(1, "big"),
            comment_id.encode(),
            b"\xD8\x01\x01\xEA\x01",
            len(post_id).to_bytes(1, "big"),
            post_id.encode(),
            b"\xF2\x01",
            len(channel_id).to_bytes(1, "big"),
            channel_id.encode(),
            b"B\x10comments-section",
        ]

        part1 = urlsafe_b64encode(b"".join(part1)).replace(b"=", b"%3D")

        params = [
            b"\xe2\xa9\x85\xb2\x02",
            (83 + 3 * len(post_id)).to_bytes(1, "big"),
            b"\x02\x12",
            len(channel_id).to_bytes(1, "big"),
            channel_id.encode(),
            b"\x1A",
            (54 + 3 * len(post_id)).to_bytes(1, "big"),
            b"\x02",
            part1,
        ]

        params = urlsafe_b64encode(b"".join(params)).decode().replace("=", "%3D")

        return params

    @staticmethod
    def from_ids(comment_id, post_id, channel_id, expire_after=0):
        fixed_comment_url = Comment.FORMAT_URLS["FIXED_COMMENT"].format(channel_id, comment_id, post_id)
        headers = {
            "x-origin": "https://www.youtube.com",
            "Referer": fixed_comment_url,
        }

        # Add authorization header
        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        c = Comment.get_fixed_comment_params(comment_id, post_id, channel_id)

        json_body = {
            "context": {
                "client": {
                    "clientName": "WEB",
                    "clientVersion": CLIENT_VERSION,
                    "originalUrl": fixed_comment_url,
                }
            },
            "continuation": c,
        }

        r = requests_cache.post(Comment.FORMAT_URLS["BROWSE_ENDPOINT"], json=json_body, expire_after=expire_after, headers=headers)

        comment_data = safely_get_value_from_key(
            r.json(), "onResponseReceivedEndpoints", 1, "reloadContinuationItemsCommand", "continuationItems", 0, "commentThreadRenderer"
        )

        if comment_data is not None:
            return Comment.from_data(
                comment_data["comment"]["commentRenderer"],
                post_id,
                channel_id,
                safely_get_value_from_key(
                    comment_data,
                    "replies",
                    "commentRepliesRenderer",
                    "contents",
                    0,
                    "continuationItemRenderer",
                    "continuationEndpoint",
                    "continuationCommand",
                    "token",
                ),
                safely_get_value_from_key(
                    comment_data,
                    "replies",
                    "commentRepliesRenderer",
                    "contents",
                    0,
                    "continuationItemRenderer",
                    "continuationEndpoint",
                    "clickTrackingParams",
                ),
                None,
                None,
            )

    @staticmethod
    def get_update_comment_params(comment_id, post_id, channel_id):
        params = [
            b"\n",
            len(comment_id).to_bytes(1, "big"),
            comment_id.encode(),
            b"*\x02\b\x00@\x01R",
            len(post_id).to_bytes(1, "big"),
            post_id.encode(),
            b"Z",
            len(channel_id).to_bytes(1, "big"),
            channel_id.encode(),
        ]

        params = urlsafe_b64encode(b"".join(params)).decode().replace("=", "%3D")

        return params

    def update_comment(self, comment_text):
        return Comment._update_comment(comment_text, comment_id=self.comment_id, post_id=self.post_id, channel_id=self.channel_id)

    @staticmethod
    def _update_comment(comment_text, update_comment_params=None, comment_id=None, post_id=None, channel_id=None):
        if update_comment_params is None:
            update_comment_params = Comment.get_update_comment_params(comment_id, post_id, channel_id)

        headers = {
            "x-origin": "https://www.youtube.com",
        }

        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        json_body = {
            "context": {
                "client": {
                    "clientName": "WEB",
                    "clientVersion": CLIENT_VERSION,
                },
            },
            "updateCommentParams": update_comment_params,
            "commentText": comment_text,
        }

        r = requests_cache.post(
            Comment.FORMAT_URLS["UPDATE_COMMENT_ENDPOINT"],
            json=json_body,
            headers=headers,
        )

        return r.json()

    @staticmethod
    def get_delete_comment_params(comment_id, post_id, channel_id):
        params = [
            b"\b\x06\x10\x07\x1A",
            len(comment_id).to_bytes(1, "big"),
            comment_id.encode(),
            b"0\x00J\x15115587043600121621724P\x00\xA8\x01\x01\xB2\x01",
            len(post_id).to_bytes(1, "big"),
            post_id.encode(),
            b"\xBA\x01",
            len(channel_id).to_bytes(1, "big"),
            channel_id.encode(),
            b"\xF0\x01\x01",
        ]

        params = urlsafe_b64encode(b"".join(params)).decode().replace("=", "%3D")

        return params

    def delete_comment(self):
        return Comment._delete_comment(comment_id=self.comment_id, post_id=self.post_id, channel_id=self.channel_id)

    @staticmethod
    def _delete_comment(delete_comment_params=None, comment_id=None, post_id=None, channel_id=None):
        if delete_comment_params is None:
            delete_comment_params = Comment.get_delete_comment_params(comment_id, post_id, channel_id)

        return Comment.perform_action(delete_comment_params)

    @staticmethod
    def get_dislike_comment_params(value, comment_id, post_id, channel_id):
        params = [
            b"\b\x04\x10\x07\x1A",
            len(comment_id).to_bytes(1, "big"),
            comment_id.encode(),
            b"0\x008",
            (not value).to_bytes(1, "big"),
            b"J\x15115587043600121621724P\x00\xA8\x01\x01\xB2\x01",
            len(post_id).to_bytes(1, "big"),
            post_id.encode(),
            b"\xBA\x01",
            len(channel_id).to_bytes(1, "big"),
            channel_id.encode(),
            b"\xF0\x01\x01",
        ]

        params = urlsafe_b64encode(b"".join(params)).decode().replace("=", "%3D")

        return params

    def set_dislike_comment(self, value=True):
        return Comment._set_dislike_comment(value, comment_id=self.comment_id, post_id=self.post_id, channel_id=self.channel_id)

    @staticmethod
    def _set_dislike_comment(value, dislike_comment_params=None, comment_id=None, post_id=None, channel_id=None):
        if dislike_comment_params is None:
            dislike_comment_params = Comment.get_dislike_comment_params(value, comment_id, post_id, channel_id)

        return Comment.perform_action(dislike_comment_params)

    @staticmethod
    def get_like_comment_params(value, comment_id, post_id, channel_id):
        params = [
            b"\b\x05\x10\x07\x1A",
            len(comment_id).to_bytes(1, "big"),
            comment_id.encode(),
            b"0\x008",
            (not value).to_bytes(1, "big"),
            b"J\x15115587043600121621724P\x00\xA8\x01\x01\xB2\x01",
            len(post_id).to_bytes(1, "big"),
            post_id.encode(),
            b"\xBA\x01",
            len(channel_id).to_bytes(1, "big"),
            channel_id.encode(),
            b"\xF0\x01\x01",
        ]

        params = urlsafe_b64encode(b"".join(params)).decode().replace("=", "%3D")

        return params

    def set_like_comment(self, value=True):
        return Comment._set_like_comment(value, comment_id=self.comment_id, post_id=self.post_id, channel_id=self.channel_id)

    @staticmethod
    def _set_like_comment(value, like_comment_params=None, comment_id=None, post_id=None, channel_id=None):
        if like_comment_params is None:
            like_comment_params = Comment.get_like_comment_params(value, comment_id, post_id, channel_id)

        return Comment.perform_action(like_comment_params)

    @staticmethod
    def perform_action(action_params):
        headers = {
            "x-origin": "https://www.youtube.com",
        }

        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        json_body = {
            "context": {
                "client": {
                    "clientName": "WEB",
                    "clientVersion": CLIENT_VERSION,
                },
            },
            "actions": [
                action_params,
            ],
        }

        r = requests_cache.post(
            Comment.FORMAT_URLS["PERFORM_COMMENT_ACTION_ENDPOINT"],
            json=json_body,
            headers=headers,
        )

        return r.json()



import json
import time
from hashlib import sha1

CLIENT_VERSION = "2.20220311.01.00"


def safely_get_value_from_key(*args, default=None):
    obj = args[0]
    keys = args[1:]

    for key in keys:
        try:
            obj = obj[key]
        except Exception:
            return default

    return obj


def safely_pop_value_from_key(*args):
    obj = args[0]
    keys = args[1:-1]

    for key in keys:
        try:
            obj = obj[key]
        except Exception:
            return None

    pop_key = args[-1]

    if pop_key in obj:
        obj.pop(pop_key)


def search_key(key, data, current_key=[]):
    found = []

    if type(data).__name__ == "dict":
        keys = list(data.keys())
    elif type(data).__name__ == "list":
        keys = list(range(len(data)))
    else:
        return []

    if key in keys:
        found.append((current_key + [key], data[key]))
        keys.remove(key)

    for k in keys:
        found += search_key(key, data[k], current_key=current_key + [k])

    return found


def save_object_to_file(obj, path):
    with open(path, "w") as f:
        f.write(json.dumps(obj, indent=4))


def get_auth_header(sapisid):
    timestring = str(int(time.time()))
    return f"SAPISIDHASH {timestring}_" + sha1(" ".join([timestring, sapisid, "https://www.youtube.com"]).encode()).hexdigest()



from .utils import (
    safely_get_value_from_key,
    save_object_to_file,
    safely_pop_value_from_key,
    search_key,
    get_auth_header,
    CLIENT_VERSION,
)
from .clean_items import (
    clean_content_text,
    clean_backstage_attachement,
)

__all__ = [
    "safely_get_value_from_key",
    "safely_pop_value_from_key",
    "save_object_to_file",
    "search_key",
    "get_auth_header",
    "clean_content_text",
    "clean_backstage_attachement",
    "CLIENT_VERSION",
]



from urllib.parse import parse_qs, unquote, urlparse
from .utils import safely_get_value_from_key as safe
from .utils import safely_pop_value_from_key as safe_pop


# lots of returned objects are full of tracking params, client data, duplicate info, etc. this sorta trims the fat.
def clean_content_text(content):
    for item in safe(content, "runs", default=[]):
        if "navigationEndpoint" in item:
            # traditional links
            if "urlEndpoint" in item["navigationEndpoint"]:
                url = item["navigationEndpoint"]["urlEndpoint"]["url"]
                # replace redirects with direct links
                if url.startswith("https://www.youtube.com/redirect"):
                    parsed_url = urlparse(item["navigationEndpoint"]["urlEndpoint"]["url"])
                    redirect_url = parse_qs(parsed_url.query)["q"][0]
                    item["urlEndpoint"] = {"url": unquote(redirect_url)}
                item.pop("navigationEndpoint")
            # hashtags
            elif "browseEndpoint" in item["navigationEndpoint"]:
                item.pop("loggingDirectives")
                safe_pop(item, "navigationEndpoint", "browseEndpoint", "params")
                item["browseEndpoint"] = item["navigationEndpoint"]["browseEndpoint"]
                item["browseEndpoint"]["url"] = item["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"]
                item.pop("navigationEndpoint")
    return content


def clean_backstage_attachement(attachment):
    if attachment:
        if "pollRenderer" in attachment:
            for choice in attachment["pollRenderer"]["choices"]:
                for value in [
                    "selectServiceEndpoint",
                    "deselectServiceEndpoint",
                    "voteRatioIfSelected",
                    "votePercentageIfSelected",
                    "voteRatioIfNotSelected",
                    "votePercentageIfNotSelected",
                ]:
                    safe_pop(choice, value)
        elif "videoRenderer" in attachment:
            safe_pop(attachment, "videoRenderer", "navigationEndpoint", "watchEndpoint", "watchEndpointSupportedOnesieConfig")
            attachment["videoRenderer"]["watchEndpoint"] = safe(attachment, "videoRenderer", "navigationEndpoint", "watchEndpoint", default={})
            attachment["videoRenderer"]["watchEndpoint"]["url"] = safe(
                attachment, "videoRenderer", "navigationEndpoint", "commandMetadata", "webCommandMetadata", "url"
            )

            for long_by_line in safe(attachment, "videoRenderer", "longBylineText", "runs", default=[]):
                long_by_line["browseEndpoint"] = long_by_line["navigationEndpoint"]["browseEndpoint"]
                long_by_line.pop("navigationEndpoint")

            for short_by_line in safe(attachment, "videoRenderer", "shortBylineText", "runs", default=[]):
                short_by_line["browseEndpoint"] = short_by_line["navigationEndpoint"]["browseEndpoint"]
                short_by_line.pop("navigationEndpoint")

            for author in safe(attachment, "videoRenderer", "ownerText", "runs", default=[]):
                author["browseEndpoint"] = author["navigationEndpoint"]["browseEndpoint"]

            for value in [
                "publishedTimeText",
                "navigationEndpoint",
                "trackingParams",
                "showActionMenu",
                "menu",
                "channelThumbnailSupportedRenderers",
                "thumbnailOverlays",
            ]:
                safe_pop(attachment, "videoRenderer", value)
        return attachment
    return None



import json

from .helpers.utils import safely_get_value_from_key


class Reply(object):
    def __init__(self, reply_id, author=None, content_text=None, vote_count=None):
        self.reply_id = reply_id
        self.author = author
        self.content_text = content_text
        self.vote_count = vote_count

    def as_json(self):
        return {"reply_id": self.reply_id, "author": self.author, "content_text": self.content_text, "vote_count": self.vote_count}

    def __str__(self):
        return json.dumps(self.as_json(), indent=4)

    def __repr__(self):
        return self.__str__()

    def get_text(self):
        if self.content_text is not None:
            return "".join([run["text"] for run in self.content_text["runs"]])
        return None

    @staticmethod
    def from_data(data):
        reply = Reply(
            data["commentId"],
            content_text=safely_get_value_from_key(data, "contentText"),
            author={
                "authorText": safely_get_value_from_key(data, "authorText"),
                "authorThumbnail": safely_get_value_from_key(data, "authorThumbnail"),
                "authorEndpoint": safely_get_value_from_key(data, "authorEndpoint", "browseEndpoint"),
                "authorIsChannelOwner": safely_get_value_from_key(data, "authorIsChannelOwner"),
                "sponsorCommentBadge": safely_get_value_from_key(data, "sponsorCommentBadge"),
            },
            vote_count=safely_get_value_from_key(data, "voteCount"),
        )

        reply.raw_data = data

        return reply



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), "r") as f:
    long_description = f.read()

setup(
    name="youtube_community_tab",
    version="0.2.3.2.1",
    description="A python3 module to handle YouTube Community Tab",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/bot-jonas/youtube-community-tab",
    package_dir={"": "src"},
    install_requires=[
        "requests_cache",
    ],
    packages=find_packages(where="src"),
    zip_safe=False,
)



from . import (
    helpers,
)

from .comment import Comment
from .community_tab import CommunityTab
from .post import Post
from .reply import Reply
from .requests_handler import requests_cache

__all__ = [
    "helpers",
    "Comment",
    "CommunityTab",
    "Post",
    "Reply",
    "requests_cache",
]



import json
import re
from requests.utils import dict_from_cookiejar
from base64 import urlsafe_b64encode

from .helpers.clean_items import clean_content_text, clean_backstage_attachement
from .helpers.utils import safely_get_value_from_key, get_auth_header, CLIENT_VERSION, search_key
from .requests_handler import requests_cache
from .comment import Comment


class Post(object):
    FORMAT_URLS = {
        "POST": "https://www.youtube.com/post/{}",
        # HARD_CODED: This key seems to be constant to everyone, IDK
        "BROWSE_ENDPOINT": "https://www.youtube.com/youtubei/v1/browse?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
        "CREATE_COMMENT_ENDPOINT": "https://www.youtube.com/youtubei/v1/comment/create_comment?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8&prettyPrint=false",
    }

    REGEX = {
        "YT_INITIAL_DATA": "ytInitialData = ({(?:(?:.|\n)*)?});</script>",
    }

    def __init__(self, post_id, channel_id=None, author=None, content_text=None, backstage_attachment=None, vote_count=None, sponsor_only_badge=None):
        self.post_id = post_id
        self.channel_id = channel_id
        self.author = author
        self.content_text = content_text
        self.backstage_attachment = backstage_attachment
        self.vote_count = vote_count
        self.sponsor_only_badge = sponsor_only_badge

        self.first = True
        self.comments = []
        self.comments_continuation_token = None
        self.click_tracking_params = None
        self.visitor_data = None
        self.session_index = "0"

    def as_json(self):
        return {
            "post_id": self.post_id,
            "channel_id": self.channel_id,
            "author": self.author,
            "content_text": self.content_text,
            "backstage_attachment": self.backstage_attachment,
            "vote_count": self.vote_count,
            "sponsor_only_badge": self.sponsor_only_badge,
        }

    @staticmethod
    def from_post_id(post_id, expire_after=0):
        headers = {"Referer": Post.FORMAT_URLS["POST"].format(post_id)}

        # Add authorization header
        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        post_url = Post.FORMAT_URLS["POST"].format(post_id)
        r = requests_cache.get(post_url, expire_after=expire_after, headers=headers)

        m = re.findall(Post.REGEX["YT_INITIAL_DATA"], r.text)
        data = json.loads(m[0])

        community_tab = data["contents"]["twoColumnBrowseResultsRenderer"]["tabs"][0]
        community_tab_items = Post.get_items_from_community_tab(community_tab)

        post_data = community_tab_items[0]["backstagePostThreadRenderer"]["post"]["backstagePostRenderer"]
        post_data["channelId"] = data["metadata"]["channelMetadataRenderer"]["externalId"]

        post = Post.from_data(post_data)
        post.get_first_continuation_token(data)
        post.get_click_tracking_params(data)
        post.visitor_data = data["responseContext"]["webResponseContextExtensionData"]["ytConfigData"]["visitorData"]
        post.session_index = str(
            safely_get_value_from_key(data, "responseContext", "webResponseContextExtensionData", "ytConfigData", "sessionIndex", default="")
        )

        return post

    def __str__(self):
        return json.dumps(self.as_json(), indent=4)

    def __repr__(self):
        return self.__str__()

    def get_thumbnails(self):
        # Returns a list of the thumbnails in different resolutions of
        # all images present in the post
        thumbnails = []

        if self.backstage_attachment is not None:
            renderer_key = list(self.backstage_attachment.keys())[0]

            if renderer_key == "videoRenderer":
                thumbnails = [self.backstage_attachment[renderer_key]["thumbnail"]["thumbnails"]]
            elif renderer_key == "backstageImageRenderer":
                thumbnails = [self.backstage_attachment[renderer_key]["image"]["thumbnails"]]
            elif renderer_key == "postMultiImageRenderer":
                thumbnails = [img["backstageImageRenderer"]["image"]["thumbnails"] for img in self.backstage_attachment[renderer_key]["images"]]
            elif renderer_key == "pollRenderer":
                print("[There is nothing implemented for polls]")
                thumbnails = []
            else:
                raise Exception("There is no implementation for renderer_key={renderer_key} yet")

        return thumbnails

    def get_first_continuation_token(self, data):
        self.comments_continuation_token = data["contents"]["twoColumnBrowseResultsRenderer"]["tabs"][0]["tabRenderer"]["content"]["sectionListRenderer"][
            "contents"
        ][1]["itemSectionRenderer"]["contents"][0]["continuationItemRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]

    def get_click_tracking_params(self, data):
        self.click_tracking_params = data["contents"]["twoColumnBrowseResultsRenderer"]["tabs"][0]["tabRenderer"]["content"]["sectionListRenderer"]["contents"][
            1
        ]["itemSectionRenderer"]["contents"][0]["continuationItemRenderer"]["continuationEndpoint"]["clickTrackingParams"]

    def load_comments(self, expire_after=0):
        headers = {"Referer": Post.FORMAT_URLS["POST"].format(self.post_id)}

        # Add authorization header
        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        if self.comments_continuation_token is None:
            try:
                r = requests_cache.get(Post.FORMAT_URLS["POST"].format(self.post_id), expire_after=expire_after, headers=headers)

                m = re.findall(Post.REGEX["YT_INITIAL_DATA"], r.text)
                data = json.loads(m[0])

                self.get_first_continuation_token(data)
                self.get_click_tracking_params(data)
                self.visitor_data = data["responseContext"]["webResponseContextExtensionData"]["ytConfigData"]["visitorData"]
                self.session_index = str(safely_get_value_from_key(data, "responseContext", "webResponseContextExtensionData", "ytConfigData", "sessionIndex"))
                self.load_comments(expire_after=expire_after)
            except Exception as e:
                print("[Some non-expected exception, probably caused by requests...]")
                raise e
        elif self.comments_continuation_token is not False:
            headers.update(
                {
                    "X-Goog-AuthUser": self.session_index,
                    "X-Origin": "https://www.youtube.com",
                    "X-Youtube-Client-Name": "1",
                    "X-Youtube-Client-Version": CLIENT_VERSION,
                }
            )

            json_body = {
                "context": {
                    "client": {
                        "clientName": "WEB",
                        "clientVersion": CLIENT_VERSION,
                        "originalUrl": Post.FORMAT_URLS["POST"].format(self.post_id),
                        "visitorData": self.visitor_data,
                    }
                },
                "continuation": self.comments_continuation_token,
                "clickTracking": {"clickTrackingParams": self.click_tracking_params},
            }

            r = requests_cache.post(Post.FORMAT_URLS["BROWSE_ENDPOINT"], json=json_body, expire_after=expire_after, headers=headers)

            data = r.json()
            if self.first:
                if "continuationItems" not in data["onResponseReceivedEndpoints"][1]["reloadContinuationItemsCommand"]:
                    # There are no comments
                    continuation_items = []
                else:
                    append = data["onResponseReceivedEndpoints"][1]["reloadContinuationItemsCommand"]
                    continuation_items = safely_get_value_from_key(append, "continuationItems", default=[])
                    self.first = False
            else:
                append = data["onResponseReceivedEndpoints"][0]["appendContinuationItemsAction"]
                continuation_items = safely_get_value_from_key(append, "continuationItems", default=[])

            self.click_tracking_params = data["trackingParams"]
            self.append_comments_from_items(continuation_items)

    def append_comments_from_items(self, items):
        there_is_no_continuation_token = True
        for item in items:
            kind = list(item.keys())[0]

            if kind == "commentThreadRenderer":
                self.comments.append(
                    Comment.from_data(
                        item[kind]["comment"]["commentRenderer"],
                        self.post_id,
                        self.channel_id,
                        safely_get_value_from_key(
                            item[kind],
                            "replies",
                            "commentRepliesRenderer",
                            "contents",
                            0,
                            "continuationItemRenderer",
                            "continuationEndpoint",
                            "continuationCommand",
                            "token",
                        ),
                        safely_get_value_from_key(
                            item[kind],
                            "replies",
                            "commentRepliesRenderer",
                            "contents",
                            0,
                            "continuationItemRenderer",
                            "continuationEndpoint",
                            "clickTrackingParams",
                        ),
                        self.visitor_data,
                        self.session_index,
                    )
                )
            elif kind == "continuationItemRenderer":
                self.comments_continuation_token = item[kind]["continuationEndpoint"]["continuationCommand"]["token"]
                there_is_no_continuation_token = False

        if there_is_no_continuation_token:
            self.comments_continuation_token = False

    def get_text(self):
        runs = safely_get_value_from_key(self.content_text, "runs", default=[])

        if self.content_text is not None:
            return "\n".join([run["text"] for run in runs])
        return None

    def get_create_comment_params(self):
        if self.channel_id is None or self.post_id is None:
            return None

        params = [
            b"*\x02\b\x00P\x01\xA2\x01",
            len(self.post_id).to_bytes(1, "big"),
            self.post_id.encode(),
            b"\xAA\x01",
            len(self.channel_id).to_bytes(1, "big"),
            self.channel_id.encode(),
        ]

        params = urlsafe_b64encode(b"".join(params)).decode().replace("=", "%3D")

        return params

    def create_comment(self, comment_text):
        headers = {
            "x-origin": "https://www.youtube.com",
        }

        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        json_body = {
            "context": {
                "client": {
                    "clientName": "WEB",
                    "clientVersion": CLIENT_VERSION,
                },
            },
            "createCommentParams": self.get_create_comment_params(),
            "commentText": comment_text,
        }

        r = requests_cache.post(
            Post.FORMAT_URLS["CREATE_COMMENT_ENDPOINT"],
            json=json_body,
            headers=headers,
        )

        try:
            data = r.json()
            comment_id = search_key("comment", data)[0][1]["commentRenderer"]["commentId"]

            return Comment.from_ids(comment_id, self.post_id, self.channel_id)
        except Exception as e:
            raise e

    @staticmethod
    def from_data(data):
        # clean the author cause it's different here for some reason
        for item in data["authorText"]["runs"]:
            item["browseEndpoint"] = item["navigationEndpoint"]["browseEndpoint"]
            item["browseEndpoint"]["url"] = item["navigationEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"]
            item.pop("navigationEndpoint")
        data["authorEndpoint"]["browseId"] = data["authorEndpoint"]["browseEndpoint"]["browseId"]
        author_url = data["authorEndpoint"]["commandMetadata"]["webCommandMetadata"]["url"]
        data["authorEndpoint"]["url"] = author_url
        for value in ["clickTrackingParams", "commandMetadata", "browseEndpoint"]:
            data["authorEndpoint"].pop(value)

        post = Post(
            data["postId"],
            channel_id=data["channelId"],
            author={
                "authorText": safely_get_value_from_key(data, "authorText"),
                "authorThumbnail": safely_get_value_from_key(data, "authorThumbnail"),
                "authorEndpoint": safely_get_value_from_key(data, "authorEndpoint"),
            },
            content_text=clean_content_text(safely_get_value_from_key(data, "contentText")),
            backstage_attachment=clean_backstage_attachement(safely_get_value_from_key(data, "backstageAttachment", default=None)),
            vote_count=safely_get_value_from_key(data, "voteCount"),
            sponsor_only_badge=safely_get_value_from_key(data, "sponsorsOnlyBadge", default=None),
        )

        post.raw_data = data

        return post

    @staticmethod
    def get_items_from_community_tab(tab):
        try:
            return tab["tabRenderer"]["content"]["sectionListRenderer"]["contents"][0]["itemSectionRenderer"]["contents"]
        except Exception as e:
            print("[Can't get the contents from the tab]")
            raise e



import json
import re
from requests.utils import dict_from_cookiejar

from .helpers.utils import safely_get_value_from_key, get_auth_header, CLIENT_VERSION, search_key
from .requests_handler import requests_cache
from .post import Post


class CommunityTab(object):
    FORMAT_URLS = {
        "COMMUNITY_TAB": "https://www.youtube.com/{}/{}/community",
        # HARD_CODED: This key seems to be constant to everyone, IDK
        "BROWSE_ENDPOINT": "https://www.youtube.com/youtubei/v1/browse?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8",
    }

    REGEX = {
        "YT_INITIAL_DATA": "ytInitialData = ({(?:(?:.|\n)*)?});</script>",
    }

    def __init__(self, channel_name):
        self.channel_name = channel_name

        self.posts_continuation_token = None
        self.click_tracking_params = None
        self.visitor_data = None
        self.session_index = "0"
        self.posts = []
        self.community_url = None
        self.channel_id = None

    def load_posts(self, expire_after=0):
        headers = {"Referer": self.community_url}

        # Add authorization header
        current_cookies = dict_from_cookiejar(requests_cache.cookies)
        if "SAPISID" in current_cookies:
            headers["Authorization"] = get_auth_header(current_cookies["SAPISID"])

        if self.posts_continuation_token is None:
            try:
                # Get posts from community tab enpoint
                self.community_url = CommunityTab.FORMAT_URLS["COMMUNITY_TAB"].format("c", self.channel_name)
                r = requests_cache.get(self.community_url, expire_after=expire_after, headers=headers)
                if r.status_code != 200:
                    self.community_url = CommunityTab.FORMAT_URLS["COMMUNITY_TAB"].format("channel", self.channel_name)
                    r = requests_cache.get(self.community_url, expire_after=expire_after, headers=headers)

                if r.status_code != 200:
                    import sys

                    print(f"[Can't get data from the channel_name: {self.channel_name}]")
                    sys.exit()

                m = re.findall(CommunityTab.REGEX["YT_INITIAL_DATA"], r.text)
                data = json.loads(m[0])

                if self.channel_id is None:
                    self.channel_id = data["metadata"]["channelMetadataRenderer"]["externalId"]

            except IndexError as e:
                print("[Can't find yt_initial_data using the regex]")
                raise e
            except json.decoder.JSONDecodeError as e:
                print("[Can't parse yt_initial_data from the regex]")
                raise e
            except Exception as e:
                print("[Some non-expected exception, probably caused by requests...]")
                raise e

            tabs = data["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]
            community_tab = CommunityTab.get_community_tab(tabs)
            community_tab_items = CommunityTab.get_items_from_community_tab(community_tab)

            self.click_tracking_params = CommunityTab.get_click_tracking_params_from_community_tab(community_tab)
            self.visitor_data = data["responseContext"]["webResponseContextExtensionData"]["ytConfigData"]["visitorData"]
            self.session_index = str(
                safely_get_value_from_key(data["responseContext"]["webResponseContextExtensionData"]["ytConfigData"], "sessionIndex", default="")
            )
            self.append_posts_from_items(community_tab_items)
        elif self.posts_continuation_token is not False:
            headers.update(
                {
                    "X-Goog-AuthUser": self.session_index,
                    "X-Origin": "https://www.youtube.com",
                    "X-Youtube-Client-Name": "1",
                    "X-Youtube-Client-Version": CLIENT_VERSION,
                }
            )

            json_body = {
                "context": {
                    "client": {"clientName": "WEB", "clientVersion": CLIENT_VERSION, "originalUrl": self.community_url, "visitorData": self.visitor_data}
                },
                "continuation": self.posts_continuation_token,
                "clickTracking": {"clickTrackingParams": self.click_tracking_params},
            }

            r = requests_cache.post(CommunityTab.FORMAT_URLS["BROWSE_ENDPOINT"], json=json_body, expire_after=expire_after, headers=headers)

            data = r.json()
            append = data["onResponseReceivedEndpoints"][0]["appendContinuationItemsAction"]
            self.click_tracking_params = data["onResponseReceivedEndpoints"][0]["clickTrackingParams"]
            self.append_posts_from_items(safely_get_value_from_key(append, "continuationItems", default=[]))

    def append_posts_from_items(self, items):
        there_is_no_continuation_token = True
        for item in items:
            kind = list(item.keys())[0]

            if kind == "backstagePostThreadRenderer":
                post_kind = list(item[kind]["post"].keys())[0]
                if post_kind == "backstagePostRenderer":
                    post_data = item[kind]["post"]["backstagePostRenderer"]
                    post_data["channelId"] = self.channel_id
                    self.posts.append(Post.from_data(post_data))
                elif post_kind == "sharedPostRenderer":
                    # TODO: parse data from item[kind]["post"]["sharedPostRenderer"]["originalPost"]["backstagePostRenderer"]
                    post_data = item[kind]["post"]["sharedPostRenderer"]
                    post_data["channelId"] = self.channel_id
                    post_data["authorText"] = post_data["displayName"]
                    post_data["authorEndpoint"] = post_data["endpoint"]
                    post_data.pop("displayName")
                    post_data.pop("endpoint")
                    self.posts.append(Post.from_data(post_data))
                else:
                    raise Exception(f"[post_kind={post_kind} is not implemented yet!]")
            elif kind == "continuationItemRenderer":
                self.posts_continuation_token = item[kind]["continuationEndpoint"]["continuationCommand"]["token"]
                there_is_no_continuation_token = False

        if there_is_no_continuation_token:
            self.posts_continuation_token = False

    @staticmethod
    def get_community_tab(tabs):
        COMMUNITY_TAB_INDEX = 3

        if len(tabs) >= COMMUNITY_TAB_INDEX + 1:
            return tabs[COMMUNITY_TAB_INDEX]
        else:
            raise Exception(f"[The community tab is expected to have index equal to {COMMUNITY_TAB_INDEX}, but len(tabs) = {len(tabs)}]")

    @staticmethod
    def get_items_from_community_tab(tab):
        try:
            return tab["tabRenderer"]["content"]["sectionListRenderer"]["contents"][0]["itemSectionRenderer"]["contents"]
        except Exception as e:
            print("[Can't get the contents from the tab]")
            raise e

    @staticmethod
    def get_click_tracking_params_from_community_tab(tab):
        try:
            return tab["tabRenderer"]["content"]["sectionListRenderer"]["trackingParams"]
        except Exception as e:
            print("[Can't get tracking params from the tab")
            raise e
